import os
import sys
import compiler
from parser import ParserError
from bike.parsing.load import load as loadast
from bike.refactor import renameFunction, renameClass, renameMethod, extractMethod
from bike.refactor.extractMethod import coords
from bike.parsing.output import save as saveast
from bike.parsing.utils import fqn_rcar, fqn_rcdr
from bike.parsing import visitor
from bike.parsing.undo import getUndoStack, UndoStackEmptyException
from bike.parsing.fastparserast import Class, Function
from bike.query.common import getScopeForLine
from bike.query.getTypeOf import getTypeOfExpr, UnfoundType
from bike.query.getReferencesToClass import getReferencesToClass
from bike.query.getReferencesToMethod import getReferencesToMethod
from bike.query.getReferencesToFunction import getReferencesToFunction
from bike.query.findDefinition import findDefinitionByCoords
from bike import log


# Returns a BRMContext object
# wants to indicate progress. The object must be callable and takes
# the following parameters:
#  - action  (string)
#  - filename (string)
def load(path):
    context = init()
    context.load(path)
    return context

def init():
    context = BRMContext_impl()
    return context

# the context object public interface
class BRMContext(object):

    def load(self, path):
        """ import some more code into bicyclerepairman """

    def save(self):
        """ save the changed files out to disk """

    def setRenameMethodPromptCallback(self, callback):
        """
        sets a callback to ask the user about method refs which brm
        can't deduce the type of. The callback must be callable, and
         take the following parameters:
          - filename
          - linenumber
          - begin column
          - end column
         (begin and end columns enclose the problematic method call)
        """
        
    def getFullyQualifiedNameOfScope(self,filename_path,line):
        """ converts line number into an fqn of the enclosing scope
        (e.g. a.b.foo.TheClass)"""

    def getTypeOfExpression(self, filename_path, line,
                           colbegin, colend):        
        """ returns a type object (which has attribute 'fqn'),
        corresponding to the type of the expression passed in """

    def renameClass(self, fqn, newname):
        """ renames a class. fqn is the fully qualified name of the
        class from the root package"""

    def renameFunction(self, fqn, newname):
        """ see renameClass """

    def renameMethod(self, fqn, newname):
        """ see renameClass """

    def renameByCoordinates(self, filename_path, line, newname):
        """ an ide friendly method which renames a class/fn/method
        pointed to by the lineno and filename"""

    def extractMethod(self, filename_path, 
                      begin_x, begin_y, 
                      end_x, end_y, 
                      methodname):
        """ extracts the region into the named method"""

    def extractFunction(self, filename_path, 
                        begin_x, begin_y, 
                        end_x, end_y, 
                        functionname):
        """ extracts the region into the named function"""

    # must be an object with a write method
    def setProgressLogger(self,logger):
        pass
        
    def undo(self):
        """ undoes the last refactoring. WARNING: this is dangerous if
        the user has modified files since the last refactoring.
        Raises UndoStackEmptyException"""

    def findReferencesByCoordinates(self, filename_path, line):
        """ given the line number of a function, class or method,
        returns a generator which finds references to it.
        """

    def findDefinitionByCoordinates(self,filename_path,line,col):        
        """ given the coordates to a reference, tries to find the
        definition of that reference """
        
class NotAPythonModuleOrPackageException: pass
class CouldntLocateASTNodeFromCoordinatesException: pass

class BRMContext_impl(BRMContext):
    
    def __init__(self):
        self.ast = None
        # Used because some refactorings delegate back to the user.
        # this flag ensures that code isnt imported during those times
        self.readyToLoadNewCode = 1 
        getUndoStack(1)  # force new undo stack


    def load(self, path):
        if not self.readyToLoadNewCode:
            return
            
        if(not path.endswith(".py")) and \
                   (not path.endswith(".py")) and \
                   os.path.isfile(path):
            raise NotAPythonModuleOrPackageException
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        class SourceVisitor:
            def visitSource(self, node):
                if node.filename == path:
                    # backup the code
                    getUndoStack().addSource(node,node.getSource())
                    getUndoStack().commitUndoFrame()
        if self.ast:
            compiler.walk(self.ast, SourceVisitor())
        self.ast = loadast(path, self.ast)

    def _getAST(self):
        return self.ast

    # returns a list of saved filenames
    def save(self):
        savedfiles = saveast(self.ast)
        return savedfiles

    def setRenameMethodPromptCallback(self, callback):
        self.promptUserClientCallback = callback

    # fqn = fully qualified name  (a.b.foo.bah)
    def renameClass(self, fqn, newname):
        #renameClass.renameClass_orig(self.ast, fqn, newname,
        renameClass.renameClass(self.ast, fqn, newname)

    def renameFunction(self, fqn, newname):
        renameFunction.renameFunction(self.ast, fqn, newname)

    def renameMethod(self, fqn, newname):
        promptcallback = None
        if hasattr(self, "promptUserClientCallback"):
            promptcallback = self._promptUser
        renameMethod.renameMethod(self.ast, fqn, newname, 
                                  promptCallback = promptcallback)

    def extractMethod(self, filename_path, 
                      begin_x, begin_y, 
                      end_x, end_y, 
                      methodname):
        ast = self._getSourceNode(filename_path)
        assert ast
        extractMethod.extractMethod(ast, coords(begin_x, begin_y), 
                                    coords(end_x, end_y), methodname)


    def extractFunction(self, filename_path, 
                      begin_x, begin_y, 
                      end_x, end_y, 
                      methodname):
        ast = self._getSourceNode(filename_path)
        assert ast
        extractMethod.extractMethod(ast, coords(begin_x, begin_y), 
                                      coords(end_x, end_y), methodname)

    def undo(self):
        getUndoStack().undo()

    def _promptUser(self, filename, lineno, colbegin, colend):
        return self.promptUserClientCallback(filename, lineno, colbegin, colend)


    # must be an object with a write method
    def setProgressLogger(self,logger):
        log.progress = logger


    # filename_path must be absolute
    def renameByCoordinates(self, filename_path, line, newname):
        self.readyToLoadNewCode = 0
        try:
            node = self._locateNode(filename_path, line)
            if node:
                if isinstance(node, Function):
                    if isinstance(node.getParent(), Class):
                        self.renameMethod(node.fqn, newname)
                    else:
                        self.renameFunction(node.fqn, newname)
                if isinstance(node, Class):
                    self.renameClass(node.fqn, newname)
            else:
                raise CouldntLocateASTNodeFromCoordinatesException()
        finally:
            self.readyToLoadNewCode = 1


    def getFullyQualifiedNameOfScope(self,filename_path,line):
        sourcenode = self._getSourceNode(filename_path)
        assert sourcenode
        scopeobj = getScopeForLine(sourcenode,line)
        return scopeobj.fqn

    def getTypeOfExpression(self, filename_path, lineno,
                           colbegin, colend):
        sourcenode = self._getSourceNode(filename_path)
        assert sourcenode
        scopeobj = getScopeForLine(sourcenode,lineno)
        assert scopeobj
        line = sourcenode.getLines()[lineno-1]
        colbegin,colend = self._reverseCoordsIfWrongWayRound(colbegin,colend)
        src = line[colbegin:colend]
        try:
            ast = compiler.parse(src,"eval")
        except ParserError:
            raise ParserError,"Could not parse string:"+src
        type = getTypeOfExpr(scopeobj,ast.node)
        if type == UnfoundType:
            return None
        return type

    def _reverseCoordsIfWrongWayRound(self, colbegin, colend):
        if(colbegin > colend):
            colbegin,colend = colend,colbegin
        return colbegin,colend


    def findDefinitionByCoordinates(self,filename_path,line,col):
        mod = self._getSourceNode(filename_path).fastparseroot
        return findDefinitionByCoords(mod,line,col)

        
    # filename_path must be absolute
    def findReferencesByCoordinates(self, filename_path, line):
        node = self._locateNode(filename_path, line)
        #print "node is",node
        if node:
            if isinstance(node, Function):
                if isinstance(node.getParent(), Class):
                    print "node is a method"
                    return getReferencesToMethod(self.ast,node.fqn)
                else:
                    print "node is a function"
                    return getReferencesToFunction(self.ast,node.fqn)
            if isinstance(node, Class):
                print "node is a class"
                return getReferencesToClass(self.ast,node.fqn)
        else:
            raise CouldntLocateASTNodeFromCoordinatesException()
        

    # fqn = fully qualified name of class
    # (see getFullyQualifiedNameOfScope())
    def getReferencesToClass(self,fqn):
        refs = []
        for ref in getReferencesToClass(self.ast,fqn):
            refs.append(ref)
        return refs

    def _getSourceNode(self, filename):
        filename = os.path.normpath(os.path.abspath(filename))
        top = self.ast
        class SourceVisitor:
            srcnode = None
            def visitSource(self, node):
                #print node.filename,filename
                if node.filename == filename:
                    self.srcnode = node
        return compiler.walk(self.ast, SourceVisitor()).srcnode


    def _locateNode(self, filename, line):
        filename = os.path.normpath(filename)
        class SourceVisitor:
            def __init__(self):
                self.result = None
            def visitSource(self, node):
                if node.filename == filename:
                    print >>log.progress,"Locating AST Node:", node.filename
                    for child in node.getClassesAndFunctions():
                        self.visit(child)

            def visitFunction(self, node):
                print >>log.progress,"Locating AST Node: Scanning",node.fqn
                if node.getStartLine() == line:
                    self.result = node
                    return
                for child in node.getChildNodes():
                    self.visit(child)

            def visitClass(self, node):
                print >>log.progress,"Locating AST Node: Scanning",node.fqn
                if node.getStartLine() == line:
                    self.result = node
                    return
                for child in node.getChildNodes():
                    self.visit(child)
                    
        result = visitor.walk(self.ast, SourceVisitor()).result
        return result
        

def _deducePackageOfFile(filename):
    package = ""
    dot = ""
    dir = os.path.dirname(filename)
    while dir != ""and \
          os.path.exists(os.path.join(dir, "__init__.py")):
        dir, dirname = os.path.split(dir)
        package = dirname+dot+package
        dot = "."
    return package
