import setpath
from compiler.ast import *
import compiler
import os
from bike.parsing.undo import getUndoStack
from bike.parsing.fastparser import fastparser
import re
import StringIO

class NoSuchPackageException:
    pass
class PackageAlreadyExistsException:
    pass
class NoSuchModuleException:
    pass
class ModuleAlreadyExistsException:
    pass

class ExtendedNode(Node):
    def _getChildren(self):
        return self._children

class PackageContainer:
    def __init__(self, packages):
        self.packages = packages

    def getPackage(self, packagename):
        packagenames = [x.name for x in self.packages]
        if packagename not in packagenames:
            raise NoSuchPackageException()
        else:
            i = packagenames.index(packagename)
            return self.packages[i]


    def addPackage(self, package):
        packagenames = [x.name for x in self.packages]
        if package.name in packagenames:
            raise PackageAlreadyExistsException()
        package.rootnode = self.rootnode
        package.setParent(self)
        if self.fqn == "":
            package.setFQN(package.name)
        else:
            package.setFQN(self.fqn+"."+package.name)
        self.packages.append(package)
            
    def hasPackage(self, packagename):
        return packagename in[x.name for x in self.packages]


class SourceContainer:
    def __init__(self, modules, rootnode = None):
        self.modules = modules
        # self.rootnode should be set by parent

    def addSource(self, source):
        if source.modulename in[x.modulename for x in self.modules]:
            raise ModuleAlreadyExistsException()
        source.setRootNode(self.rootnode)
        source.setParent(self)
        self.modules.append(source)


    def getSource(self, modulename):
        modnames = [x.modulename for x in self.modules]
        try:
            return self.modules[modnames.index(modulename)]
        except ValueError:
            raise NoSuchModuleException()

    def hasSource(self, modulename):
        return modulename in[x.modulename for x in self.modules]


class PackageAndSourceContainer(PackageContainer, SourceContainer):
    # jesus! I need some inspiration in the class naming dept
    pass

class Root(ExtendedNode, PackageAndSourceContainer):
    # Always the root of the extended ast.
    nodes['root'] = 'Root'

    def __init__(self, packages = None, modules = None, pythonpath = None):

        # this is to get round an optimisation which reuses an empty list as a default arg
        #  unfortunately the client of this method may fill that list, so it's not empty
        if not packages:
            packages = []
        if not modules:
            modules = []
        if not pythonpath:
            pythonpath = []
        PackageContainer.__init__(self, packages)
        SourceContainer.__init__(self, modules)
        self.fqn = ""
        self.pythonpath = pythonpath
        self._children = ('top', packages, modules, pythonpath)
        for package in packages:
            package.setParent(self)
        for mod in modules:
            mod.setParent(self)
        self.rootnode = self

    def __repr__(self):
        return "Root(%s,%s,%s)"%self._children[1:]

    def getChildNodes(self):
        nodes = []
        if self.packages is not None: nodes+=self.packages
        if self.modules is not None: nodes+=self.modules
        return tuple(nodes)


class Package(ExtendedNode, PackageAndSourceContainer):
    nodes['package'] = 'Package'

    def __init__(self, path, name, rootnode = None):
        if rootnode is not None:
            self.rootnode = rootnode
        PackageContainer.__init__(self, [])
        SourceContainer.__init__(self, [])
        self.path = path
        self.name = name
        self._children = self.packages, self.modules

    def setFQN(self, fqn):
        self.fqn = fqn

    def setParent(self, scope):
        self.parentScope = scope

    def getParent(self):
        return self.parentScope

    def getChildNodes(self):
        nodes = []
        if self.packages is not None: nodes+=self.packages
        if self.modules is not None: nodes+=self.modules
        return tuple(nodes)

    def __repr__(self):
        return "Package(%s,%s,%s,%s)"%(self.path, self.name, self.packages, self.modules)


wordRE = re.compile("\w+")
class Source(ExtendedNode):
    nodes['Source'] = 'Source'

    def createFromString(filename, modulename, src):
        s = StringIO.StringIO(src)
        return Source(filename,modulename,src)
    createFromString = staticmethod(createFromString)

    def createFromFile(filename,modulename):
        try:
            f = file(filename)
            return Source(filename,modulename,f.read())
        finally:
            f.close()
    createFromFile = staticmethod(createFromFile)
    
    def __init__(self, filename, modulename, src):
        self._lazyloading = 1

        if os.path.isabs(filename):
            self.filename = filename
        else:
            self.filename = os.path.abspath(filename)
        self.modulename = modulename

        self.load(src)


    def load(self, source):
        # fastparser ast
        self.fastparseroot = fastparser(source,self.modulename)
        self.fastparseroot.setSourceNode(self)
        self._lines = source.splitlines(1)

        # tokens
        self._tokens = None
        self._tokenized = 0

        self._module = None

        self._modified = 0
        self._tokensModified = 0
        self.sourcenode = self
        self._source = source

    def __repr__(self):
        #return "Source(%s,%s,%s)"%('source', self.filename, self.modulename, self.getModule())
        return "Source(%s,%s,%s)"%('source', self.filename, self.getModule())

    def getChildren(self):
        return self.getChildNodes()

    #def getChildNodes(self):
    #    return tuple([self.getModule()])

    def getChildNodes(self):
        return self.getClassesAndFunctions()

    def isModified(self):
        return self._modified

    # remove me!
    def refreshTokens(self):
        self.syncSourceWithTokens()

    def setModified(self):
        self._modified = 1

    def setUpToDate(self):
        self._modified = 0

    def setTokensModified(self):
        self._tokensModified = 1
        self.setModified()

    def tokensModified(self):
        return self._tokensModified

    def syncSourceWithTokens(self):
        self._source = self.getSourceAsString()
        self._tokensModified = 0
        self.setModified()

    def setRootNode(self, rootnode):
        self.rootnode = rootnode

    def reload(self, source):
        self.load(source)
        self.setParent(self.parentScope)

    def setFQN(self, fqn):
        self.fqn = fqn

    def setParent(self, scope):
        self.parentScope = scope
        
        if self.parentScope.fqn == "":
            self.setFQN(self.modulename)
        else:
            self.setFQN(self.parentScope.fqn+"."+self.modulename)

        # set the fastparserroot stuff
        self.fastparseroot.setParent(scope)
        #self.fastparseroot.setFQN(self.fqn)
        

    def getSourceAsString(self):
        tokens = self._getTokens_internal()
        return tokensToString(tokens.begin(), tokens.end())

    def getSource(self):
        if(self.tokensModified()):
            self.syncSourceWithTokens()
        if self.isModified():
            self._source = "".join(self.getLines())
        return self._source

    def getClassesAndFunctions(self):
        return self.fastparseroot.getChildNodes()


    # TODO: rename me!
    def getFlattenedListOfFastParserASTNodes(self):
        return self.fastparseroot.getFlattenedListOfChildNodes()
        
    def containsString(self,string):
        if self.getSource().find(string) != -1:
            #print self.getLines()
            for line in self.getLines():
                if line.find(string) != -1:
                    words = wordRE.findall(line)
                    if string in words:
                        return 1
        return 0

    def getLines(self):
        return self._lines

# describes an instance of something (e.g. instance of a class)
class Instance(compiler.ast.Node):
    def __init__(self, type):
        assert type is not None
        self._type = type

    def getType(self):
        return self._type

    def __str__(self):
        return "Instance(%s)"%(self.getType())





# Local Variables:          
# mode:python               
# indent-tabs-mode: nil     
# py-indent-offset: 4   
# py-smart-indentation: nil 
# End:                      

