#!/usr/bin/env python
import setpath
import unittest
import compiler
import os

from bike import testdata
from bike.testutils import*
from bike.mock import Mock

from load import _get_modpkg_path
from load import*
import load as loadmodule

class TestGetFilesForName(unittest.TestCase):
    def testGetFilesForName_recursivelyReturnsFilesInBreadthFirstOrder(self):
        createPackageStructure("pass", "pass")

        files = getFilesForName(pkgstructureBasedir)
        for f in files:
            assert f in \
                  [os.path.join(pkgstructureBasedir, '__init__.py'), 
                    os.path.join(pkgstructureBasedir, 'foo.py'), 
                    os.path.join(pkgstructureChilddir, '__init__.py'), 
                    os.path.join(pkgstructureChilddir, 'bah.py')]
        removePackageStructure()

    def testGetFilesForName_globsStars(self):
        createPackageStructure("pass", "pass")
        assert getFilesForName(os.path.join(pkgstructureBasedir, "fo*")) == [os.path.join(pkgstructureBasedir, 'foo.py')]
        removePackageStructure()

class TestLoad(unittest.TestCase):
    def testLoad_returnsEmptyRootNodeIfNoFilesMatched(self):
        root = load("doesntexist")
        assert root.modules == []
        assert root.packages == []


    def testLoad_returnsASTContainingDepthFirstStructure(self):
        createPackageStructure("pass", "pass")
        class ASTVisitor:
            def __init__(self):
                self.visited = []
            def visitSource(self, node, scope = 0):
                self.visited.append(node.filename)
                self.visited.append(node.modulename)
                #self.visit(node.getModule())
            #def visitPass(self, node, scope = 0):
            #    self.visited.append('pass')
        root = load(pkgstructureBasedir)
        removePackageStructure()
        a = compiler.walk(root, ASTVisitor())
        for item in a.visited:
            assert item in \
              [os.path.join(os.getcwd(), pkgstructureChilddir, '__init__.py'), 
                '__init__', 
                os.path.join(os.getcwd(), pkgstructureChilddir, 'bah.py'), 
                'bah', 
                os.path.join(os.getcwd(), pkgstructureBasedir, '__init__.py'), 
                '__init__', 
                os.path.join(os.getcwd(), pkgstructureBasedir, 'foo.py'), 
                'foo']

    def testLoad_returnsRootNodeWithSourceIfOnlyOneFileImported(self):
        writeTmpTestFile("pass")
        root = load(tmpfile)
        class ASTVisitor:
            def __init__(self): self.visited = []
            def visitSource(self, node, scope = 0):
                self.visited.append(node.filename)
        a = compiler.walk(root, ASTVisitor())
        assert a.visited == [os.path.abspath(tmpfile)]
        deleteTmpTestFile()

    def testLoad_doesntThrowExceptionCalledTwiceOnDifferentTreesSections(self):
        createPackageStructure("ClassWithRefToOwnMethod", "ClassNestedInFunction")
        root = load(pkgstructureBasedir)
        root = load(pkgstructureChilddir, root)


    def testLoad_reloadsModuleIfAlreadyInAST(self):
        createPackageStructure("pass", "pass")
        root = load(pkgstructureBasedir)
        removePackageStructure()
        createPackageStructure("pass", testdata.class_and_method)
        root = load(pkgstructureFile2, root)
        removePackageStructure()
        assert getASTNode(root, "Class").name == "a"

    #if given an multi-element path to root package
    #(e.g. foo/bah/root, doesnt include other directories in package
    #structure
    def testLoad_doesntIncludeNonPackageDirectoriesPreceedingRootInPackageStructure(self):
        createPackageStructure("pass", "pass")
        root = load(os.path.join(pkgstructureBasedir, 'b', '..', '..', 'a'))
        removePackageStructure()
        class ASTVisitor:
            def __init__(self): self.visited = []
            def visitPackage(self, node):
                self.visited.append(node.name)
                for i in node.getChildNodes():
                    self.visit(i)
        a = compiler.walk(root, ASTVisitor())
        assert len(a.visited) == 2
        assert a.visited[0] == 'a'
        assert a.visited[1] == 'b'


    def test_includesParentPackagesWhenLoadingModule(self):
        createPackageStructure("pass", "pass")
        root = load(pkgstructureFile2)
        removePackageStructure()
        class ASTVisitor:
            def __init__(self): self.visited = []
            def visitPackage(self, node):
                self.visited.append(node.name)
                for i in node.getChildNodes():
                    self.visit(i)
        a = compiler.walk(root, ASTVisitor())
        assert len(a.visited) == 2
        assert a.visited[0] == 'a'
        assert a.visited[1] == 'b'
        assert getASTNode(root, "Source").fqn == "a.b.bah"


    def test_doesntIncludeNonPackageSubDirectoriesInAST(self):
        createPackageStructure("pass","pass")
        os.makedirs(os.path.join(pkgstructureBasedir,"z"))
        writeFile(os.path.join(pkgstructureBasedir,"z","shouldntloadmodule.py"),"pass")
        root = load(pkgstructureBasedir)        
        os.remove(os.path.join(pkgstructureBasedir,"z","shouldntloadmodule.py"))
        os.removedirs(os.path.join(pkgstructureBasedir,"z"))
        removePackageStructure()
        assert getMatchingASTNode(root,"Module",name="shouldntloadmodule") == None

    def test_includesModulesInRootDirectoryEvenIfDirectoryIsNotAPackage(self):
        os.makedirs(pkgstructureBasedir)
        writeFile(os.path.join(pkgstructureBasedir,"mymodule.py"),"pass")
        root = load(pkgstructureBasedir)
        os.remove(os.path.join(pkgstructureBasedir,"mymodule.py"))
        os.removedirs(pkgstructureBasedir)
        assert getMatchingASTNode(root,"Source",fqn="mymodule") != None
        



class TestGetRootDirectory(unittest.TestCase):
    def test_returnsParentDirectoryIfFileNotInPackage(self):
        try:
            # this doesnt have __init__.py file, so
            # isnt package
            os.makedirs("a")
            writeFile(os.path.join("a", "foo.py"), "pass")
            dir = loadmodule.getRootDirectory(os.path.join("a", "foo.py"))
            assert dir == "a"
        finally:
            os.remove(os.path.join("a", "foo.py"))
            os.removedirs(os.path.join("a"))

    def test_returnsFirstNonPackageParentDirectoryIfFileInPackage(self):
        try:
            os.makedirs(os.path.join("root", "a", "b"))
            writeFile(os.path.join("root", "a", "__init__.py"), "# ")
            writeFile(os.path.join("root", "a", "b", "__init__.py"), "# ")
            writeFile(os.path.join("root", "a", "b", "foo.py"), "pass")
            dir = loadmodule.getRootDirectory(os.path.join("root", "a", "b", "foo.py"))
            assert dir == "root"
        finally:
            os.remove(os.path.join("root", "a", "__init__.py"))
            os.remove(os.path.join("root", "a", "b", "__init__.py"))
            os.remove(os.path.join("root", "a", "b", "foo.py"))
            os.removedirs(os.path.join("root", "a", "b"))

    def test_returnsFirstNonPackageParentDirectoryIfPathIsAPackage(self):
        try:
            os.makedirs(os.path.join("root", "a", "b"))
            writeFile(os.path.join("root", "a", "__init__.py"), "# ")
            writeFile(os.path.join("root", "a", "b", "__init__.py"), "# ")
            writeFile(os.path.join("root", "a", "b", "foo.py"), "pass")
            dir = loadmodule.getRootDirectory(os.path.join("root", "a", "b"))
            assert dir == "root"
        finally:
            os.remove(os.path.join("root", "a", "__init__.py"))
            os.remove(os.path.join("root", "a", "b", "__init__.py"))
            os.remove(os.path.join("root", "a", "b", "foo.py"))
            os.removedirs(os.path.join("root", "a", "b"))

    def test_returnsDirIfDirIsTheRootDirectory(self):
        try:
            os.makedirs(os.path.join("root", "a", "b"))
            writeFile(os.path.join("root", "a", "__init__.py"), "# ")
            writeFile(os.path.join("root", "a", "b", "__init__.py"), "# ")
            writeFile(os.path.join("root", "a", "b", "foo.py"), "pass")
            dir = loadmodule.getRootDirectory("root")
            assert dir == "root"
        finally:
            os.remove(os.path.join("root", "a", "__init__.py"))
            os.remove(os.path.join("root", "a", "b", "__init__.py"))
            os.remove(os.path.join("root", "a", "b", "foo.py"))
            os.removedirs(os.path.join("root", "a", "b"))

if __name__ == "__main__":
    unittest.main()
