# getTypeOf(scope,fqn) and getTypeOfExpr(scope,ast)

from bike.parsing.fastparserast import Class, Function, Module
from bike.parsing.extended_ast import Source, Package, Root, NoSuchModuleException, NoSuchPackageException, Instance
from bike.parsing.parserutils import generateLogicalLines, makeLineParseable,splitLogicalLines, makeLineParseable
from bike.parsing import visitor
import re
import compiler

# used if an assignment exists, but cant find type
# e.g. a = SomeFunctionNotLoaded()
# (as opposed to 'None' if cant find an assignment)
class UnfoundType: pass


# name is the fqn of the reference, scope is the scope ast object from
# which the question is being asked.
# returns an fastparser-ast object representing the type
# or None if type not found
def getTypeOf(scope, fqn):
    #if not isinstance(scope,Instance):
    #    print "getTypeOf",scope.fqn,fqn

    if isinstance(scope,UnfoundType):
        return UnfoundType()

    if "."in fqn:
        car = fqn.split(".")[0]
        cdr = ".".join(fqn.split(".")[1:])

        newscope = getTypeOf(scope, car)
        #print "newscope is ",newscope
        if newscope is not None:
            return getTypeOf(newscope, cdr)

    if isinstance(scope, Root) or isinstance(scope, Package):
        try:
            child = scope.getSource(fqn).fastparseroot
        except NoSuchModuleException:
            #print fqn
            try:
                child = scope.getPackage(fqn)
            except NoSuchPackageException:
                child = None
        if child:
            return child
        else:
            if isinstance(scope, Package):
                if scope.hasSource("__init__"):
                    type = getImportedType(scope.getSource("__init__").fastparseroot, fqn)
                    if type:
                        return type
                # maybe fqn is absolute (src node delegates to package)
                return getTypeOf(scope.rootnode, fqn)

    elif isinstance(scope,Instance):  # scope is a class instance reference

        # if it's an instance, search instance class for method names
        # then search methods for assignments with self.foo getattrs
        
        # first search the instance type for members

        matches = [c for c in scope.getType().getChildNodes() if c.name == fqn]
        if matches != []:
            return matches[0]

        # then search the functions of the class for self.fqn assignments
        for child in scope.getType().getChildNodes():
            if not isinstance(child,Function):
                continue
            res = scanScopeAST(child,fqn,SelfAttributeAssignmentVisitor(child,fqn))
            if res is not None:
                return res
        
    else:  # scope is source, class or function

        if fqn == "self" and isinstance(scope,Function) and \
               isinstance(scope.getParent(),Class):
            return Instance(scope.getParent())

        #print "scope is ",scope.fqn
        matches = [c for c in scope.getChildNodes()if c.name == fqn]
        if matches != []:
            return matches[0]

        type = scanScopeSourceForType(scope, fqn)
        if type != None:
            return type
        
        type = getImportedType(scope, fqn)   # try imported types
        if type != None:
            return type

        parentScope = scope.getParent()
        while isinstance(parentScope,Class):
            # don't search class scope, since this is not accessible except
            # through self
            parentScope = parentScope.getParent()
        return getTypeOf(parentScope, fqn)


wordRE = re.compile("\w+")
def isWordInLine(word, line):
    if line.find(word) != -1:
        words = wordRE.findall(line)
        if word in words:
            return 1
    return 0

# scans the import statements in the scope to see if it imports fqn
def getImportedType(scope, fqn):
    lines = scope.module.getSourceNode().getLines()
    for lineno in scope.getImportLines():
        logicalline = generateLogicalLines(lines[lineno-1:]).next()
        #print "logicalline:",logicalline
        logicalline = makeLineParseable(logicalline)
        ast = compiler.parse(logicalline)
        class ImportVisitor:
            def __init__(self):
                self.match = None
            def visitImport(self, node):
                for name, alias in node.names:
                    if alias is None:
                        # will be handled automatically because
                        # typechecking delegates to parents
                        pass
                    else:
                        if alias == fqn:
                            self.match = getTypeOf(scope.getParent(),name)

            def visitFrom(self, node):
                if node.names[0][0] == '*': # e.g. from foo import *
                    if not "."in fqn:
                        module = getTypeOf(scope.getParent(), node.modname)
                        if module:
                            self.match = getTypeOf(module, fqn)
                else:
                    for name, alias in node.names:
                        if alias == fqn:
                            module = getTypeOf(scope.getParent(), node.modname)
                            if module:
                                self.match = getTypeOf(module, name)
                        elif alias is None and name == fqn:
                            module = getTypeOf(scope.getParent(), node.modname)
                            if module:
                                self.match = getTypeOf(module, name)
                                
        match = visitor.walk(ast, ImportVisitor()).match
        if match:
            return match
        #else loop

class TypeNotSupportedException:
    def __init__(self,msg):
        self.msg = msg

    def __str__(self):
        return self.msg

# attempts to evaluate the type of the expression
def getTypeOfExpr(scope, ast):
    #print "ast",ast
    if isinstance(ast, compiler.ast.Name):
        return getTypeOf(scope, ast.name)

    elif isinstance(ast, compiler.ast.Getattr):
        expr = getTypeOfExpr(scope, ast.expr)
        #if isinstance(expr,Instance):
        #    expr = expr.getType()
        #    print "expr is",expr.fqn
        if expr is not None:
            attrnametype = getTypeOf(expr, ast.attrname)
            #print "attrnametype",attrnametype
            return attrnametype
        return None

    elif isinstance(ast, compiler.ast.CallFunc):
        node = getTypeOfExpr(scope,ast.node)
        if isinstance(node,Class):
            return Instance(node)
        elif isinstance(node,Function):
            return getReturnTypeOfFunction(node)
    else:
        raise TypeNotSupportedException, "Evaluation of "+str(ast)+" not supported"

def getReturnTypeOfFunction(function):
    return scanScopeAST(function,"return",ReturnTypeVisitor(function))
    


# does parse of scope sourcecode to deduce type
def scanScopeSourceForType(scope, fqn):
    return scanScopeAST(scope,fqn,AssignmentVisitor(scope,fqn))


# scans for lines containing keyword, and then runs the visitor over
# the parsed AST for that line
def scanScopeAST(scope,keyword,astvisitor):
    lines = scope.getLinesNotIncludingThoseBelongingToChildScopes()
    src = ''.join(lines)
    match = None
    for line in splitLogicalLines(src):
        if isWordInLine(keyword, line):
            doctoredline = makeLineParseable(line)
            ast = compiler.parse(doctoredline)
            match = visitor.walk(ast,astvisitor).getMatch()
            if match:
                return match
    return match
    

class AssignmentVisitor:
    def __init__(self,scope,targetName):
        self.match=None
        self.scope = scope
        self.targetName = targetName

    def getMatch(self):
        return self.match

    def visitAssign(self,node):
        if isinstance(node.expr,compiler.ast.CallFunc):
            for assnode in node.nodes:
                if isinstance(assnode,compiler.ast.AssName) and \
                   assnode.name == self.targetName:
                    #print "Found",assnode
                    #print "expr=",getTypeOfExpr(self.scope,node.expr)
                    self.match = getTypeOfExpr(self.scope,node.expr)
                    if self.match is None:
                        self.match = UnfoundType()
                    
                    

class SelfAttributeAssignmentVisitor:
    def __init__(self,scope,targetName):
        self.match=None
        self.scope = scope
        self.targetName = targetName

    def getMatch(self):
        return self.match

    def visitAssign(self,node):
        if isinstance(node.expr,compiler.ast.CallFunc):
            for assnode in node.nodes:
                if isinstance(assnode,compiler.ast.AssAttr) and \
                   isinstance(assnode.expr,compiler.ast.Name) and \
                   assnode.expr.name == "self" and \
                   assnode.attrname == self.targetName:
                    self.match = getTypeOfExpr(self.scope,node.expr)
                    #print "here!",self.match.getType().fqn


class ReturnTypeVisitor:
    def __init__(self,fn):
        self.match=None
        self.fn = fn

    def getMatch(self):
        return self.match
    
    def visitReturn(self,node):
        try:
            self.match = getTypeOfExpr(self.fn,node.value)
        except TypeNotSupportedException, ex:
            pass


