#!/usr/bin/env python
import setpath
import unittest
import os
from bike import testdata
from bike.testutils import *
from bike.query.getReferencesToClass import *
from bike.parsing.fastparserast import Module

class TestGetReferencesToClass(unittest.TestCase):
    def test_returnsEmptyListIfNoReferences(self):
        src = trimLines("""
        class MyClass:
            pass
        a = TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        self.assertEqual([x for x in getReferencesToClass(root,"mymodule.MyClass")],[])

    def test_findsSimpleReferenceInSameModule(self):
        src = trimLines("""
        class TheClass:
            pass
        a = TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        refs = [x for x in getReferencesToClass(root,"mymodule.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath("mymodule.py"))
        self.assertEqual(refs[0].modulefqn,"mymodule")
        self.assertEqual(refs[0].lineno,3)
        self.assertEqual(refs[0].colno,4)
        self.assertEqual(refs[0].confidence,100)

    def test_findsReferencesInModuleWhichImportsClass(self):
        src = trimLines("""
        import b.bah.TheClass
        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]

        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,1)
        self.assertEqual(refs[0].colno,13)
        self.assertEqual(refs[0].confidence,100)

        self.assertEqual(refs[1].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[1].modulefqn,"a.foo")
        self.assertEqual(refs[1].lineno,3)
        self.assertEqual(refs[1].colno,14)
        self.assertEqual(refs[0].confidence,100)


    def test_findsReferenceInModuleWhichImportsClassWithFrom(self):
        src = trimLines("""
        from b.bah import TheClass
        def foo():
            a = TheClass()
            a.theMethod()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,1)
        self.assertEqual(refs[0].colno,18)
        self.assertEqual(refs[0].confidence,100)

        self.assertEqual(refs[1].filename,os.path.abspath(os.path.join("a/foo.py")))
        self.assertEqual(refs[1].modulefqn,"a.foo")
        self.assertEqual(refs[1].lineno,3)
        self.assertEqual(refs[1].colno,8)
        self.assertEqual(refs[0].confidence,100)


    def test_findsReferenceInModuleWhichImportsClassWithFromAndAlias(self):
        src = trimLines("""
        from b.bah import TheClass as MyTheClass
        def foo():
            a = MyTheClass()
            a.theMethod()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,1)
        self.assertEqual(refs[0].colno,18)
        self.assertEqual(refs[0].confidence,100)


    def test_findsReferenceInModuleWhichImportsClassWithImportAs(self):
        src = trimLines("""
        import b.bah.TheClass as MyTheClass
        def foo():
            a = MyTheClass()
            a.theMethod()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,1)
        self.assertEqual(refs[0].colno,13)
        self.assertEqual(refs[0].confidence,100)

    def test_findsReferenceInModuleWhichImportsClassWithFromImportStar(self):
        src = trimLines("""
        from b.bah import *
        a = TheClass()
        a.theMethod()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,2)
        self.assertEqual(refs[0].colno,4)
        self.assertEqual(refs[0].confidence,100)

    def test_findsReferenceInModuleWhichImportsClassWithFromImportStar2(self):
        src = trimLines("""
        from a.b.bah import *
        a = TheClass()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,2)
        self.assertEqual(refs[0].colno,4)
        self.assertEqual(refs[0].confidence,100)


    def test_findsClassReferenceInInstanceCreation(self):
        src = trimLines("""
        class TheClass:
            def theMethod(self): pass
        TheClass().theMethod()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        refs = [x for x in getReferencesToClass(root,"a.foo.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,3)
        self.assertEqual(refs[0].colno,0)
        self.assertEqual(refs[0].confidence,100)

    
    def test_findsClassReferenceInInstanceCreationWithFQN(self):
        src = trimLines("""
        import b.bah.TheClass
        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,1)
        self.assertEqual(refs[0].colno,13)
        self.assertEqual(refs[0].confidence,100)

        self.assertEqual(refs[1].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[1].modulefqn,"a.foo")
        self.assertEqual(refs[1].lineno,3)
        self.assertEqual(refs[1].colno,14)
        self.assertEqual(refs[0].confidence,100)

    """   Can't do this! - needs logic which would massively complicate things    
    def test_doesntfindReferenceInModuleWhichDoesntImportClass(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceButNotImported
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs,[])
    """


    def test_findsReferenceInClassBases(self):
        src =trimLines("""
        from b.bah import TheClass
        class DerivedClass(TheClass):
            pass
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[1].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[1].modulefqn,"a.foo")
        self.assertEqual(refs[1].lineno,2)
        self.assertEqual(refs[1].colno,19)
        self.assertEqual(refs[1].confidence,100)
        

    
    def test_findsReferenceInMultiLineImportStatement(self):
        src =trimLines("""
        from b.bah import foo, \\
                  TheFooBah, TheClass, TheBastard,SomethingElse
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,2)
        self.assertEqual(refs[0].colno,21)
        self.assertEqual(refs[0].confidence,100)

    def test_findsReferenceWhenModulenameSameAsClassMethodName(self):
        # asserts that brm doesnt search class scope after not finding name
        # in method scope (since class scope is invisible unless called on 'self'
        src =trimLines("""
        from a.b import bah
        class baz:
            def bah(self):
                print bah.TheClass
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,4)
        self.assertEqual(refs[0].colno,18)
        self.assertEqual(refs[0].confidence,100)


    def test_doesntBarfOnFromImportStarWhenNameIsInFromClause(self):
        src = trimLines("""
        from a.b.bah.TheClass import *
        a = TheClass()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testdata.TheClass,
                                  "a.b.bah.TheClass")
        refs = [x for x in getReferencesToClass(root,"a.b.bah.TheClass")]


class TestGetScopeForLine(unittest.TestCase):
    def test_worksWithFunctionScope(self):
        src = trimLines("""
        class a:
            def foo():
                pass
        """)
        node = createAST(src)
        self.assertEqual(getScopeForLine(node,3).name,"foo")
        
    def test_worksWithModuleScope(self):
        src = trimLines("""
        class TheClass:
            pass
        a = TheClass()
        """)
        node = createAST(src)
        assert isinstance(getScopeForLine(node,3),Module)
        
if __name__ == "__main__":
    unittest.main()
