#!/usr/bin/env python
import setpath
import unittest
from renameMethod import*
import compiler

from bike import testdata

from bike.testutils import*

from bike.parsing.load import load
#from bike.ui import test_bikecli

# Generic tests. These tests are designed to be run in the context of a ui
#  and in a package hierarchy structure
class RenameMethodTests:

    # template method
    def renameMethod(self, src, classfqn, methodname, newname):
        ''' override to call the real renameMethod and return
        the resultant src as a string'''
        pass

    def test_renamesTheMethod(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod():
                pass
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName():
                pass
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_doesntRenameMethodOfSameNameOnOtherClasses(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod():
                pass
        class b:
            def theMethod():
                pass
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName():
                pass
        class b:
            def theMethod():
                pass
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_doesntRenameOtherMethodsOfSameClass(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod(self):
                a=b
            def aMethod(self):
                pass
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName(self):
                a=b
            def aMethod(self):
                pass
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_renamesMethodWhenClassNestedInFunction(self):
        srcBefore=trimLines("""
        def theFunction():
            class TheClass:
                def theMethod(self):
                    pass
        """)
        srcAfter=trimLines("""
        def theFunction():
            class TheClass:
                def newName(self):
                    pass
        """)
        src = self.renameMethod(srcBefore,"theFunction.TheClass.theMethod",
                                "newName")
        self.assertEqual(srcAfter,src)

    def test_doesntBarfOnInheritanceHierarchies(self):
        srcBefore=trimLines("""
        from b.bah import DifferentClass
        class TheClass(foo.bah):
            def theMethod(self):
                pass
        """)
        self.renameMethod(srcBefore,"TheClass.theMethod","newName")

    def test_renamesMethodWhenMethodCallFromOtherMethodInSameClass(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod(self):
                pass
            def anotherMethod(self):
                self.theMethod()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName(self):
                pass
            def anotherMethod(self):
                self.newName()
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_doesntBarfOnNestedClasses(self):
        srcBefore=trimLines("""
        class TheClass:
            class AnotherClass:
                pass
            def theMethod(self):
                pass
        """)
        self.renameMethod(srcBefore,"TheClass.theMethod","newName")

    def test_renamesMethodWhenBaseClassesArentInAST(self):
        srcBefore=trimLines("""
        class TheClass(notInAst):
            def theMethod(self):
                pass
        """)
        srcAfter=trimLines("""
        class TheClass(notInAst):
            def newName(self):
                pass
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_renamesMethodInRelatedClasses(self):
        srcBefore=trimLines("""
        class root:
            def theMethod():
                pass

        class a(root):
            def theMethod():
                pass
    
        class b(root):
            pass

        class TheClass(b):
            def theMethod():
                pass
        """)
        srcAfter=trimLines("""
        class root:
            def newName():
                pass

        class a(root):
            def newName():
                pass
    
        class b(root):
            pass

        class TheClass(b):
            def newName():
                pass
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)


    def test_renameMethodDoesntBarfOnNoneAsDefaultArgToMethod(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod(self, root, flist, stack=None):
                pass
        """)
        self.renameMethod(srcBefore,"TheClass.theMethod","newName")



class RenameMethodTests_ImportsClass:
    def test_renamesMethodOnDerivedClassInstance(self):
        srcBefore = trimLines("""
        from b.bah import TheClass as BaseClass
        
        class DerivedClass(BaseClass):
            pass

        class DerivedDerivedClass(DerivedClass):
            def theMethod(self):
                print 'hello'
        """)
        srcAfter = trimLines("""
        from b.bah import TheClass as BaseClass
        
        class DerivedClass(BaseClass):
            pass

        class DerivedDerivedClass(DerivedClass):
            def newName(self):
                print 'hello'
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)
        
class RenameMethodReferenceTests:
    # Generic tests. These tests are designed to be run in the context of a ui
    #  and in a package hierarchy structure

    def test_doesntBarfWhenConfrontedWithComplexReturnTypes(self):
        src = trimLines("""
        import a
        class TheClass:
            def theMethod(self):
                 pass
                 
        def bah():
            return a[35]
            
        b = bah()
        b.theMethod()
        """)
        self.renameMethod(src, "TheClass.theMethod", "newName")

    def test_doesntbarfWhenCallMadeOnInstanceReturnedFromFnCall(self):
        srcBefore=trimLines("""
        from foo import e
        class TheClass:
            def theMethod(self):
                pass
        ast = e().f(src)
        """)
        self.renameMethod(srcBefore,"TheClass.theMethod","newName")

    def test_doesntStackOverflowOnRecursiveFunctions(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod(self):
                pass

        def foo(a):
            return foo(a)
        """)
        self.renameMethod(srcBefore,"TheClass.theMethod","newName")

    def test_renamesMethodReferenceOfInstanceCreatedInParentScopeAfterFunction(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod(self):
                pass
        a = TheClass()
        def foo():
            a.theMethod()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName(self):
                pass
        a = TheClass()
        def foo():
            a.newName()
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)
 
    def test_renamesMethodReferenceOfInstanceObtainedByCallingFunction(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod():
                pass
        def foo():
            b = TheClass()
            return b
        a = foo()
        a.theMethod()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName():
                pass
        def foo():
            b = TheClass()
            return b
        a = foo()
        a.newName()
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_renamesMethodReferenceOfInstanceCreatedInAnotherFunction(self):

        srcBefore=trimLines("""
        class TheClass:
            def theMethod():
                pass
        def bah():
            return TheClass()
        def foo():
            a = bah()
            a.theMethod()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName():
                pass
        def bah():
            return TheClass()
        def foo():
            a = bah()
            a.newName()
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_renamesMethodReferenceOfInstanceCreatedInSubsequentFunction(self):
        srcBefore = trimLines("""
        class TheClass:
            def theMethod():
                pass
        class NotTheClass:
            def theMethod():
                pass
            
        def foo():
            a = bah()
            a.theMethod()
            
        def bah():
            return TheClass()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName():
                pass
        class NotTheClass:
            def theMethod():
                pass
            
        def foo():
            a = bah()
            a.newName()
            
        def bah():
            return TheClass()
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_renamesMethodReferenceOnInstanceThatIsAnAttributeOfSelf(self):
        srcBefore = trimLines("""
        class TheClass:
            def theMethod(self):
                pass
        
        class AnotherClass:
            def __init__(self):
                self.a = TheClass()
            def anotherFn(self):
                self.a.theMethod()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName(self):
                pass
        
        class AnotherClass:
            def __init__(self):
                self.a = TheClass()
            def anotherFn(self):
                self.a.newName()
        """)
        src = self.renameMethod(srcBefore,"TheClass.theMethod","newName")
        self.assertEqual(srcAfter,src)

    def test_doesntBarfOnGetattrThatItCantDeduceTypeOf(self):
        srcBefore=trimLines("""
        class TheClass:
            def theMethod(self):
                pass
        a = TheClass
        
        a.b.bah = 3
        """)
        self.renameMethod(srcBefore,"TheClass.theMethod","newName")


class RenameMethodReferenceTests_ImportsClass:

    def test_renamesReferenceOfClassImportedAsAnotherName(self):
        srcBefore=trimLines("""
        import b.bah.TheClass as MyTheClass        
        def foo():
            a = MyTheClass()
            a.theMethod()
        """)
        srcAfter=trimLines("""
        import b.bah.TheClass as MyTheClass        
        def foo():
            a = MyTheClass()
            a.newName()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)

    def test_renamesReferenceWhenObjectCreationAndReferenceInModuleScope(self):
        srcBefore=trimLines("""
        from b.bah import TheClass
        a = TheClass()
        a.theMethod()
        """)
        srcAfter=trimLines("""
        from b.bah import TheClass
        a = TheClass()
        a.newName()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)


    def test_renamesReferenceWhenObjectCreatedInSameFunctionAsReference(self):
        srcBefore=trimLines("""
        import b.bah.TheClass
        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        srcAfter=trimLines("""
        import b.bah.TheClass
        def foo():
            a = b.bah.TheClass()
            a.newName()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)

    def test_doesntrenameDifferentMethodReferenceWhenObjectCreatedInSameScope(self):
        srcBefore=trimLines("""
        import b.bah.TheClass
        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.differentMethod", "newName")
        self.assertEqual(srcBefore,src)

    def test_doesntrenameMethodReferenceWhenDifferentObjectCreatedInSameScope(self):
        srcBefore=trimLines("""
        import b.bah.TheClass
        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.DifferentClass.theMethod",
                                            "newName")
        self.assertEqual(srcBefore,src)

    def test_renamesReferenceOfImportedClass(self):
        srcBefore=trimLines("""
        import b.bah.TheClass
        
        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        srcAfter=trimLines("""
        import b.bah.TheClass
        
        def foo():
            a = b.bah.TheClass()
            a.newName()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)

    def test_doesntRenameReferenceOfDifferentImportedClass(self):
        srcBefore=trimLines("""
        import b.bah.DifferentClass
        
        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.DifferentClass.theMethod",
                                           "newName")
        self.assertEqual(srcBefore,src)

    def test_renamesReferenceOfClassImportedWithFromClause(self):
        srcBefore=trimLines("""
        from b.bah import TheClass
        
        def foo():
            a = TheClass()
            a.theMethod()
        """)
        srcAfter=trimLines("""
        from b.bah import TheClass
        
        def foo():
            a = TheClass()
            a.newName()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)

    def test_doesntrenameReferenceOfClassImportedWithDifferentAsClause(self):
        srcBefore = trimLines("""
        from b.bah import TheClass as MyClass
        
        def foo():
            a = TheClass()
            a.theMethod()
        """)

        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcBefore,src)

    def test_renamesReferenceOfClassImportedWithFromFooImportStar(self):
        srcBefore=trimLines("""
        from b.bah import *
        a = TheClass()
        a.theMethod()
        """)
        srcAfter=trimLines("""
        from b.bah import *
        a = TheClass()
        a.newName()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)

    def test_renamesMethodReferenceOfInstanceCreatedInParentScope(self):
        srcBefore=trimLines("""
        from b.bah import TheClass
        a = TheClass()
        def foo():
            a.theMethod()
        """)
        srcAfter=trimLines("""
        from b.bah import TheClass
        a = TheClass()
        def foo():
            a.newName()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)

    def test_doesntRenameMethodWhenObjectCreatedInChildScopeToMethodReference(self):
        srcBefore = trimLines("""
        from b.bah import TheClass
        a = AnotherClass()
        def foo():
            a = TheClass()
        a.theMethod()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcBefore,src)

    def test_renamesReferenceOnDerivedClassInstance(self):
        srcBefore=trimLines("""
        import b
        class DerivedClass(b.bah.TheClass):
            pass
        class DerivedDerivedClass(DerivedClass):
            pass
        theInstance = DerivedDerivedClass()
        theInstance.theMethod()
        """)
        srcAfter=trimLines("""
        import b
        class DerivedClass(b.bah.TheClass):
            pass
        class DerivedDerivedClass(DerivedClass):
            pass
        theInstance = DerivedDerivedClass()
        theInstance.newName()
        """)
        src = self.renameMethod(srcBefore, "a.b.bah.TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)



# tests that cover stuff not renamed automatically
# (I.e. are renamed after user manually expresses desire to do so)
class RenameMethodAfterPromptTests:
    def test_renamesReferenceWhenMethodCallDoneOnInstanceCreation(self):

        srcBefore=trimLines("""
        class TheClass:
            def theMethod(self): pass
        TheClass().theMethod()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName(self): pass
        TheClass().newName()
        """)
        src = self.renameMethod(srcBefore, "TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)


    def test_renamesReferenceInMiddleOfBiggerCompoundCall(self):
        srcBefore = trimLines("""
        class TheClass:
            def theMethod(self): return AnotherClass()
        TheClass().theMethod().anotherMethod()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName(self): return AnotherClass()
        TheClass().newName().anotherMethod()
        """)
        src = self.renameMethod(srcBefore, "TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)

def createNestedAST(src1, src2):
    createPackageStructure(src1, src2)
    ast = load(pkgstructureBasedir)
    removePackageStructure()
    return ast


class TestRenameMethodWithSingleModule(RenameMethodTests, RenameMethodReferenceTests, unittest.TestCase):
    # template method 
    def renameMethod(self, src, fqn, newname):
        root = createSourceNodeAt(Root(), src, "foo")
        renameMethod(root,"foo."+fqn,newname)
        return getASTNode(root,"Source").getSource()


class TestRenameMethodWithDirectoryStructure(RenameMethodTests, RenameMethodReferenceTests, unittest.TestCase):

    def renameMethod(self, src, fqn, newname):
        root = createNestedAST("pass", src)
        renameMethod(root,"a.b.bah."+fqn,newname)
        return getMatchingASTNode(root,"Source",fqn="a.b.bah").getSource()


class TestRenameMethodReferenceWithDirectoryStructure(RenameMethodTests_ImportsClass, RenameMethodReferenceTests_ImportsClass, unittest.TestCase):

    def renameMethod(self, src, fqn, newname):
        root = createNestedAST(src,testdata.TheClass)
        renameMethod(root,fqn,newname)
        return getMatchingASTNode(root,"Source",fqn="a.foo").getSource()


class TestRenameMethodReferenceTwice(RenameMethodTests_ImportsClass, RenameMethodReferenceTests_ImportsClass, unittest.TestCase):

    def renameMethod(self, src, fqn, newname):
        root = createNestedAST(src,testdata.TheClass)
        renameMethod(root,fqn,"anotherName")
        renameMethod(root,fqn_rcdr(fqn)+".anotherName",newname)
        return getMatchingASTNode(root,"Source",fqn="a.foo").getSource()


class TestRenameMethodStuffCorrectlyAfterPromptReturnsTrue(unittest.TestCase, 
                                                      RenameMethodAfterPromptTests):

    class MockPromptCallback(Mock):
        def callback(self, filename, line, colbegin, colend):
            return 1


    def renameMethod(self, src, fqn, newname):
        root = createNestedAST(src,testdata.TheClass)
        renameMethod(root,"a.foo."+fqn,newname,
                     promptCallback = self.MockPromptCallback().callback)
        return getMatchingASTNode(root,"Source",fqn="a.foo").getSource()


class TestDoesntRenameMethodIfPromptReturnsFalse(unittest.TestCase):
    def callback(self, filename, line, colbegin, colend):
        return 0

    def renameMethod(self, src, fqn, newname):
        root = createSourceNodeAt(Root(), src, "foo")
        renameMethod(root,"foo."+fqn,newname,
                     promptCallback = self.callback)
        return getMatchingASTNode(root,"Source").getSource()


    def test_doesntRenameMethodIfPromptReturnsFalse(self):
        srcBefore = trimLines("""
        class TheClass:
            def theMethod(self):
                pass
        b = TheClass()
        b.theMethod()
        a = someFunction()
        a.theMethod()
        """)
        srcAfter=trimLines("""
        class TheClass:
            def newName(self):
                pass
        b = TheClass()
        b.newName()
        a = someFunction()
        a.theMethod()
        """)
        src = self.renameMethod(srcBefore, "TheClass.theMethod", "newName")
        self.assertEqual(srcAfter,src)

if __name__ == "__main__":
    unittest.main()
